use vulkan_sys::prelude::*;

static mut INSTANCE_FN_HANDLES: Option<VkInstanceHandles> = None;

pub fn is_instance_fns_initialized() -> bool {
    unsafe { INSTANCE_FN_HANDLES.is_some() }
}

pub fn instance_fns() -> &'static VkInstanceHandles {
    unsafe { INSTANCE_FN_HANDLES.as_ref().unwrap() }
}

pub fn set_instance_fns(handles: VkInstanceHandles) {
    unsafe { INSTANCE_FN_HANDLES = Some(handles) };
}

pub fn replace_device_proc_addr(proc_addr: PFN_vkGetDeviceProcAddr) {
    unsafe {
        INSTANCE_FN_HANDLES
            .as_mut()
            .unwrap()
            .instance_functions
            .vkGetDeviceProcAddr = proc_addr
    };
}

pub struct VkInstanceHandles {
    pub instance: VkInstance,

    pub static_functions: StaticFunctions,
    pub entry_functions: EntryFunctions,

    pub instance_functions: InstanceFunctions,
    pub instance_wsi_functions: InstanceWSIFunctions,
    pub physical_device_properties2_functions: PhysicalDeviceProperties2Functions,
    pub debug_report_callback_functions: DebugReportCallbackFunctions,
    pub debug_utils_messenger_functions: DebugUtilsMessengerFunctions,
}

impl VkInstanceHandles {
    pub fn load_instance(
        static_functions: StaticFunctions,
        entry_functions: EntryFunctions,
        instance: VkInstance,
    ) -> Self {
        Self {
            instance_functions: InstanceFunctions::new(&static_functions, instance),
            instance_wsi_functions: InstanceWSIFunctions::new(&static_functions, instance),
            physical_device_properties2_functions: PhysicalDeviceProperties2Functions::new(
                &static_functions,
                instance,
            ),
            debug_report_callback_functions: DebugReportCallbackFunctions::new(
                &static_functions,
                instance,
            ),
            debug_utils_messenger_functions: DebugUtilsMessengerFunctions::new(
                &static_functions,
                instance,
            ),

            instance,
            static_functions,
            entry_functions,
        }
    }
}

macro_rules! cmp_instance_fn {
    ($name:ident, {$([$fns:ident, $fn_name:ident],)*}) => {
        match $name {
            $(
                stringify!($fn_name) => return unsafe {
                    mem::transmute(
                        instance_fns()
                            .$fns
                            .$fn_name,
                    )
                },
            )*

            _ => ()
        }

    }
}
